package test;

import java.io.*;

public class Generator {

    private static final boolean GP_OUTPUT = true;

    private static final boolean USEMULTI = false; // REQUIRES MultiJava compiler!
    private static final boolean USEINSTANCE = false;
    private static final boolean USEVISITOR = true;
    private static final boolean USEDEDICATED = false; 
    private static final boolean USERUNABOUT = true; 

    private static final int defITER = 1; // how many iterations (total)
    private static final int defVISITOR_COUNT = 1000; // how many visitors?
    private static final boolean defFLAT = true; // flat class hierarchy or maximum-depth?
    private static final int defREPEAT = 100; // how often to repeat in the same VM (to run hot)


    public static void main(String[] args) 
	throws IOException {
	for (int k=1;k<=200;k++) {
	    System.out.println("Generating code with " + 
			       k + (defFLAT?" FLAT":" DEEP") +
			       " classes and " + defVISITOR_COUNT + 
			       " visitors running " + defITER +
			       " iterations");
	    String fname = "src/test/Driver" + k + ".java";
	    PrintStream stream 
		= new PrintStream(new FileOutputStream(new File(fname)));
	    new Generator(stream,
			  k,
			  defVISITOR_COUNT,
			  defITER/k,
			  defREPEAT,
			  defFLAT).dump("Driver"+k);
	}
    }

    /* ******************** Generator instance ****************** */

    private final java.io.PrintStream out;
 
    private final int VISITEE_COUNT;
    private final int ITER;
    private final int VISITOR_COUNT;
    private final boolean FLAT;
    private final int REPEAT;   

    Generator(PrintStream o,
	      int visitees,
	      int visitors,
	      int iter,
	      int repeat,
	      boolean flat) {
	this.out = o;
	this.VISITEE_COUNT = visitees;
	this.ITER = iter;
	this.VISITOR_COUNT = visitors;
	this.FLAT = flat;
	this.REPEAT = repeat;
    }

    /**
     * Dump i-th visitee class with basename "A" in flat/deep hierarchy.
     **/
    private  void dumpVisitee(int i, 
			      boolean flat) {
	if (flat || (i == 0)) {
	    out.println("public static class A"+i+" implements A {\n");
	if (USEVISITOR)
  	    out.println(" public void accept(Visitor v) { v.visit(this); }\n");
	if (USEDEDICATED)
	    for (int k=0;k<VISITOR_COUNT;k++)
		out.println(" public int dedicated"+k+"() { return " + i + "; }\n");
	    out.println("} // end of A" + i);
	} else {
	    out.println("public static class A" + i +
			" extends A" + (i-1)+ " implements A {\n" +
			" public void accept(Visitor v) { v.visit(this); }\n");
	    for (int k=0;k<VISITOR_COUNT;k++)
		out.println(" public int dedicated"+k+"() { return " + i + "; }\n");
	    out.println("} // end of A" + i);
	}
    }

    private  void dumpBaseClasses(int maxVisitors) {
	// base visitee
	out.println("  public interface A {\n");
	if (USEDEDICATED)
  	  for (int k=0;k<maxVisitors;k++)
	     out.println("int dedicated"+k+"();\n");
	if (USEVISITOR) 
	  out.println("    public void accept(Visitor v);");
        out.println("  }\n");
    }

    private  void dumpVisitors(int maxVisitees,
			       int maxVisitors) {
 	// abst. visitor
	out.println("public static abstract class Visitor {\n");
	for (int i=0;i<maxVisitees;i++)
	    out.println("  public abstract void visit(A" + i + " a);");
	out.println("} // end of Visitor");

	for (int k=0;k<maxVisitors;k++) {
	    out.println("public static class Visitor"+k+
			       " extends Visitor {\n  long sum = 0;");
	    for (int i=0;i<maxVisitees;i++)
		out.println("  public void visit(A" + i + " a) { sum += " + i + ";}");
	    out.println("} // end of Visitor"+k);
	}
    }

    private  void dumpRunabouts(int maxVisitees,
				int maxVisitors) {
	// runabout
	for (int k=0;k<maxVisitors;k++) {
	    out.println("public static class SumRunabout"+k+" extends Runabout {\n"+
			       "  long sum = 0;\n"+
			       "  public SumRunabout"+k+"() {super(); }");
	    for (int i=0;i<maxVisitees;i++)
		out.println("  public void visit(A" + i + " a) { sum += " + i + ";}");
	    out.println("} // end of SumRunabout"+k);
	}
    }

    private void dumpMultiJava(int maxVisitees,
			       int maxVisitors) {
	for (int k=0;k<maxVisitors;k++) {	    
	    out.println("public static class SumMultiJava"+k+" {\n  long sum = 0;");
	    out.println("  public void visit(A a) { throw new Error(\"impossible\");}");
	    for (int i=0;i<maxVisitees;i++)
		out.println("  public void visit(A@A" + i + " a) { sum += " + i + ";}");
	    out.println("} // end of SumMultiJava"+k);
	    } 
    }

    void dumpInnerClasses() {
 	dumpBaseClasses(VISITOR_COUNT);
	// visitees
 	 for (int i=0;i<VISITEE_COUNT;i++)
	    dumpVisitee(i, FLAT);
        if (USEVISITOR)
	  dumpVisitors(VISITEE_COUNT,
	  	     VISITOR_COUNT);
	if (USERUNABOUT)
	  dumpRunabouts(VISITEE_COUNT,
		      VISITOR_COUNT);
	if (USEMULTI) 
	    dumpMultiJava(VISITEE_COUNT,
			  VISITOR_COUNT);
    }

    private void createVisitees() {
	out.println("    final A[] a = new A[" + VISITEE_COUNT+"];");
	for (int i=0;i<VISITEE_COUNT;i++)
	    out.println("    a["+i+ "] = new A"+i+"();");
    }

    private void dumpLoop(String callee,
			  int loop) {
        out.println("    "+callee+"();");
	out.println("    int[] results"+callee+" = new int[" + REPEAT + "];");
 	out.println("    for (int repeat=0;repeat<"+REPEAT+";repeat++)");
	out.println("      results"+callee+"[repeat] = " + callee + "();");
	out.println("    long sum"+callee+" = 0;\n"+
		    "    long sqsum"+callee+" = 0;\n"+
		    "    long min"+callee+" = results"+callee+"[0];\n"+
		    "    long max"+callee+" = results"+callee+"[0];\n"+
		    "    for (int i=0;i<"+REPEAT+";i++) {\n"+
		    "        sum"+callee+" += results"+callee+"[i];\n"+
		    "        sqsum"+callee+" += results"+callee+"[i] * results"+callee+"[i];\n"+
		    "        if (results"+callee+"[i] > max"+callee+")\n"+
		    "    	max"+callee+" = results"+callee+"[i];\n"+
		    "        if (results"+callee+"[i] < min"+callee+")\n"+
		    "    	min"+callee+" = results"+callee+"[i];\n"+		
		    "    }");
	if (GP_OUTPUT) 
	    out.println("    System.out.println(\""+callee+" "+ VISITEE_COUNT+ " \" + (sum"+callee+"/"+REPEAT+") + \n"+
			"      		    \"  \" + (Math.sqrt((sqsum"+callee+" - (sum"+callee+"*sum"+callee+"/"+REPEAT+"))/"+(REPEAT-1)+")) +\n"+
			"    	            \"  \" + min"+callee+" +\n"+
			"                       \"  \" + max"+callee+");");
	else
	    out.println("    System.out.println(\""+callee+": AVG: \" + (sum"+callee+"/"+REPEAT+") + \n"+
			"      		    \" SDV: \" + (Math.sqrt((sqsum"+callee+" - (sum"+callee+"*sum"+callee+"/"+REPEAT+"))/"+(REPEAT-1)+")) +\n"+
			"    	            \" MIN: \" + min"+callee+" +\n"+
			"                       \" MAX: \" + max"+callee+");");
    }

    private void startTimer() {
  	out.println("    long start = System.currentTimeMillis();");
    }

    private void stopTimer() {
 	out.println("    long end = System.currentTimeMillis();\n" +
		    "    return (int) (end - start);");
    }

    private void dumpDedicated() {
	out.println("  static int dedicated() {");
	createVisitees();
	startTimer();
	out.println( "    long sum = 0;\n" +
		     "    for (int i=0;i<"+ITER+";i++)\n"+
		     "      for (int j=0;j<"+VISITEE_COUNT+";j++) {\n");
	for (int k=0;k<VISITOR_COUNT;k++)
	    out.println("           sum += a[j].dedicated"+k+"();");
	out.println("      }\n");
	stopTimer();
 	out.println("  }");
    }
    
    private void dumpVisitor() {
	out.println("  static int visitors() {");
        startTimer();
	for (int k=0;k<VISITOR_COUNT;k++)
	    out.println("      Visitor"+k+" visitor"+k+" = new Visitor"+k+"();");
	stopTimer();
	out.println("  }");
    }

    private void dumpRunabouts() {
	out.println("  static int runabouts() {");
	startTimer();
        for (int k=0;k<VISITOR_COUNT;k++)
          out.println("    SumRunabout"+k+" runabout"+k+" = new SumRunabout"+k+"();\n");
	stopTimer();
  	out.println("  }");
    }

    private void dumpInstanceof() {
 	out.println("  static int instanceofs() {");
	createVisitees();
	startTimer();
	out.println( "    long sumInstanceof = 0;\n" +
		     "    for (int i=0;i<"+ITER+";i++)\n"+
		     "      for (int j=0;j<"+VISITEE_COUNT+";j++) {\n"+
		     "        A tmp = a[j];");
	
	for (int k=0;k<VISITOR_COUNT;k++)
	    for (int j=VISITEE_COUNT-1;j>=0;j--)
		out.println("    if (tmp instanceof A"+j+") sumInstanceof += " + j + "; else");
	out.println("        throw new Error(\"Illegal call\");");	
	out.println("      }\n");
	stopTimer();	
  	out.println("  }");
    }

    private void dumpMulti() {
	out.println("  static int multijava() {");
	createVisitees();
	for (int k=0;k<VISITOR_COUNT;k++)
	    out.println("    SumMultiJava"+k+" multisum"+k+" = new SumMultiJava"+k+"();\n");
	startTimer();
	out.println("    for (int i=0;i<"+ITER+";i++)\n"+
		    "      for (int j=0;j<"+VISITEE_COUNT+";j++) {\n");
	for (int k=0;k<VISITOR_COUNT;k++)
	    out.println("        multisum"+k+".visit(a[j]);");
	out.println("      }\n");	
	stopTimer();	
  	out.println("  }");
    }  

    public void dump(String name) {
	
	out.println("package test;\nimport ovm.util.Runabout;");
	out.println("public class "+name+" { \n");
	dumpInnerClasses();
	
	// main
	out.println("\n\n  public static void main(String[] args) {");	
	if (USEDEDICATED) 
	    dumpLoop("dedicated", REPEAT);
	if (USEVISITOR)
	    dumpLoop("visitors", REPEAT);
	if (USERUNABOUT)
	    dumpLoop("runabouts", REPEAT);
	if (USEINSTANCE)
	    dumpLoop("instanceofs", REPEAT);
	if (USEMULTI)
	    dumpLoop("multijava", REPEAT);
	out.println("}\n\n"); // end main
	
	// main.dedicated:
        if (USEDEDICATED) 
	    dumpDedicated();
        if (USEVISITOR) 
	    dumpVisitor();
	if (USERUNABOUT)
	    dumpRunabouts();
	if (USEINSTANCE) 
	    dumpInstanceof();
      	if (USEMULTI) 
	    dumpMulti();	
	
	// Driver end
	out.println("\n\n} // end Driver\n\n");
    }

} // end of Generator
