package test;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import ovm.util.Runabout;

/**
 * Testcase for Runabout.
 * @author jv, cg
 */
public class TestRunabout extends TestCase {
    static int counter;

    public void setUp() {
        counter = 0;
    }

    public static void main(String args[]) {
        junit.textui.TestRunner.run(suite());
	//new TestRunabout("A").testArray();
    }

    public static Test suite() {
        return new TestSuite(TestRunabout.class);
    }

    public TestRunabout(String s) {
        super(s);
    }

    public void test1() {
        Runabout tc = new T1();
        tc.visitAppropriate("Hello World");
        tc.visitAppropriate(new A());
        assertTrue(counter == 2);
    }
    public class T1 extends Runabout {
        public void visit(String s) {
            if (s == "Hello World")
                counter++;
        }
        public void visit(A a) {
            counter++;
        }
    }

    public void test2() {
        Runabout tc = new T2();
        tc.visitAppropriate(new A());
        assertTrue(counter == 1);
    }
    public class T2 extends Runabout {
        public void visit(AA a) {
            counter++;
        }
    }

    public void test3() {
        Runabout tc = new T3();
        tc.visitAppropriate(new A());
        assertTrue(counter == 1);
    }
    public class T3 extends Runabout {
        public void visit(AA a) {
            counter--;
        }
        public void visit(A a) {
            counter++;
        }
    }

    /**
     * Test that in an ambiguous case an interface is selected
     * randomly
     */
    public void test4() {
        Runabout tc = new T4();
	try {
	    tc.visitAppropriate(new A());
	    assertTrue(false); // should have excepted
	} catch (Runabout.RunaboutException re) {
	    assertTrue(true);
	}
    }
    public class T4 extends Runabout {
        public void visit(AA a) {
            counter--;
        }
        public void visit(AB a) {
            counter++;
        }
    }

    /**
     * Array test with subtyping
     */
    public void testArray() {
        Runabout tc = new TArray();
	tc.visitAppropriate(new String[5]);
	assertTrue(counter == 1);
	tc.visitAppropriate(new Object[3]);
	assertTrue(counter == 2);
    }
    public class TArray extends Runabout {
        public void visit(Object[] a) {
            counter++;
        }
    }

    /**
     * Array test with subtyping
     */
    public void testMultiArray() {
        Runabout tc = new TMArray();
	tc.visitAppropriate(new String[5][3][5]);
	assertTrue(counter == 1);
    }
    public class TMArray extends Runabout {
        public void visit(Object[][][] a) {
            counter++;
        }
    }

    /**
     * Array test with subtyping.
     * String[][][] is a 'subtype' of Object[][]!
     * (Since String[] is an Object).
     */
    public void testCrazyMultiArray() {
        Runabout tc = new TCMArray();
	tc.visitAppropriate(new String[5][3][5]);
	assertTrue(counter == 1);
    }
    public class TCMArray extends Runabout {
        public void visit(Object[][] a) {
            counter++;
        }
    }

    /**
     * Array test without subtyping
     */
    public void testArray2() {
        Runabout tc = new TArray2();
	tc.visitAppropriate(new String[5]);
	assertTrue(counter == 1);
    }
    public class TArray2 extends Runabout {
        public void visit(String[] a) {
            counter++;
        }
    }

    public interface AA {
    }
    public interface AB {
    }
    public class A implements AA, AB {
    }

} // end of TestRunabout
